use crate::constants;
use crate::controllers::{HttpResult, LibVirtReq};
use crate::libvirt_lib_structures::XMLUuid;
use crate::libvirt_rest_structures::nw_filter::NetworkFilter;
use actix_web::{HttpResponse, web};

#[derive(serde::Serialize, serde::Deserialize)]
pub struct NetworkFilterID {
    uid: XMLUuid,
}

/// Create a new nw filter
pub async fn create(client: LibVirtReq, req: web::Json<NetworkFilter>) -> HttpResult {
    let network = match req.0.rest2lib() {
        Ok(d) => d,
        Err(e) => {
            log::error!("Failed to extract network filter info! {e}");
            return Ok(HttpResponse::BadRequest()
                .json(format!("Failed to extract network filter info! {e}")));
        }
    };

    if constants::BUILTIN_NETWORK_FILTER_RULES.contains(&network.name.as_str()) {
        return Ok(HttpResponse::ExpectationFailed()
            .json("Builtin network filter rules shall not be modified!"));
    }

    let uid = match client.update_network_filter(req.0, network).await {
        Ok(u) => u,
        Err(e) => {
            log::error!("Failed to update network filter! {e}");
            return Ok(HttpResponse::InternalServerError()
                .json(format!("Failed to update network filter! {e}")));
        }
    };

    Ok(HttpResponse::Ok().json(NetworkFilterID { uid }))
}

/// Get the list of network filters
pub async fn list(client: LibVirtReq) -> HttpResult {
    let networks = match client.get_full_network_filters_list().await {
        Err(e) => {
            log::error!("Failed to get the list of network filters! {e}");
            return Ok(HttpResponse::InternalServerError()
                .json(format!("Failed to get the list of networks! {e}")));
        }
        Ok(l) => l,
    };

    let networks = networks
        .into_iter()
        .map(|n| NetworkFilter::lib2rest(n).unwrap())
        .collect::<Vec<_>>();

    Ok(HttpResponse::Ok().json(networks))
}

/// Get the information about a single network filter
pub async fn get_single(client: LibVirtReq, req: web::Path<NetworkFilterID>) -> HttpResult {
    let nwfilter = NetworkFilter::lib2rest(client.get_single_network_filter(req.uid).await?)?;
    Ok(HttpResponse::Ok().json(nwfilter))
}

/// Get the XML source description of a single network filter
pub async fn single_src(client: LibVirtReq, req: web::Path<NetworkFilterID>) -> HttpResult {
    let xml = client.get_single_network_filter_xml(req.uid).await?;
    Ok(HttpResponse::Ok().content_type("application/xml").body(xml))
}

/// Update the information about a single network filter
pub async fn update(
    client: LibVirtReq,
    path: web::Path<NetworkFilterID>,
    body: web::Json<NetworkFilter>,
) -> HttpResult {
    let mut network = match body.0.rest2lib() {
        Ok(n) => n,
        Err(e) => {
            log::error!("Failed to extract network filter info! {e}");
            return Ok(HttpResponse::BadRequest()
                .json(format!("Failed to extract network filter info!\n${e}")));
        }
    };
    network.uuid = Some(path.uid);

    if constants::BUILTIN_NETWORK_FILTER_RULES.contains(&network.name.as_str()) {
        return Ok(HttpResponse::ExpectationFailed()
            .json("Builtin network filter rules shall not be modified!"));
    }

    if let Err(e) = client.update_network_filter(body.0, network).await {
        log::error!("Failed to update network filter! {e}");
        return Ok(HttpResponse::InternalServerError()
            .json(format!("Failed to update network filter!\n${e}")));
    }

    Ok(HttpResponse::Ok().json("Network filter updated"))
}

/// Delete a network filter
pub async fn delete(client: LibVirtReq, path: web::Path<NetworkFilterID>) -> HttpResult {
    // Prevent deletion of default rules
    let network = client.get_single_network_filter(path.uid).await?;
    if constants::BUILTIN_NETWORK_FILTER_RULES.contains(&network.name.as_str()) {
        return Ok(HttpResponse::ExpectationFailed()
            .json("Builtin network filter rules shall not be deleted!"));
    }

    client.delete_network_filter(path.uid).await?;

    Ok(HttpResponse::Ok().json("Network deleted"))
}
