use crate::libvirt_lib_structures::XMLUuid;
use crate::libvirt_lib_structures::nwfilter::{
    NetworkFilterRefXML, NetworkFilterRuleProtocolAllXML, NetworkFilterRuleProtocolArpXML,
    NetworkFilterRuleProtocolIpvx, NetworkFilterRuleProtocolLayer4, NetworkFilterRuleProtocolMac,
    NetworkFilterRuleXML, NetworkFilterXML,
};
use crate::libvirt_rest_structures::LibVirtStructError;
use crate::libvirt_rest_structures::LibVirtStructError::{
    NetworkFilterExtraction, StructureExtraction,
};
use crate::utils::net_utils;
use lazy_regex::regex;
use std::net::{Ipv4Addr, Ipv6Addr};

pub fn is_var_def(var: &str) -> bool {
    lazy_regex::regex!("^\\$[a-zA-Z0-9_]+$").is_match(var)
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilterName(pub String);

impl NetworkFilterName {
    pub fn is_valid(&self) -> bool {
        regex!("^[a-zA-Z0-9-_]+$").is_match(&self.0)
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilterMacAddressOrVar(pub String);

impl NetworkFilterMacAddressOrVar {
    pub fn is_valid(&self) -> bool {
        is_var_def(&self.0) || net_utils::is_mac_address_valid(&self.0)
    }
}

impl From<&String> for NetworkFilterMacAddressOrVar {
    fn from(value: &String) -> Self {
        Self(value.to_string())
    }
}

fn extract_mac_address_or_var(
    n: &Option<NetworkFilterMacAddressOrVar>,
) -> anyhow::Result<Option<String>> {
    if let Some(mac) = n {
        if !mac.is_valid() {
            return Err(NetworkFilterExtraction(format!(
                "Invalid mac address or variable! {}",
                mac.0
            ))
            .into());
        }
    }

    Ok(n.as_ref().map(|n| n.0.to_string()))
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilterIPOrVar<const V: usize>(pub String);
pub type NetworkFilterIPv4OrVar = NetworkFilterIPOrVar<4>;

impl<const V: usize> NetworkFilterIPOrVar<V> {
    pub fn is_valid(&self) -> bool {
        is_var_def(&self.0)
            || match V {
                4 => net_utils::is_ipv4_address_valid(&self.0),
                6 => net_utils::is_ipv6_address_valid(&self.0),
                _ => panic!("Unsupported IP version!"),
            }
    }
}

impl<const V: usize> From<&String> for NetworkFilterIPOrVar<V> {
    fn from(value: &String) -> Self {
        if V != 4 && V != 6 {
            panic!("Unsupported IP version!");
        }
        Self(value.to_string())
    }
}

fn extract_ip_or_var<const V: usize>(
    n: &Option<NetworkFilterIPOrVar<V>>,
) -> anyhow::Result<Option<String>> {
    if let Some(ip) = n {
        if !ip.is_valid() {
            return Err(NetworkFilterExtraction(format!(
                "Invalid IPv{V} address or variable! {}",
                ip.0
            ))
            .into());
        }
    }

    Ok(n.as_ref().map(|n| n.0.to_string()))
}

fn extract_ip_mask<const V: usize>(n: Option<u8>) -> anyhow::Result<Option<u8>> {
    if let Some(mask) = n {
        if !net_utils::is_mask_valid(V, mask) {
            return Err(NetworkFilterExtraction(format!("Invalid IPv{V} mask! {mask}")).into());
        }
    }

    Ok(n)
}

fn extract_nw_filter_comment(n: &Option<String>) -> anyhow::Result<Option<String>> {
    if let Some(comment) = n {
        if comment.len() > 256 || comment.contains('\"') || comment.contains('\n') {
            return Err(NetworkFilterExtraction(format!("Invalid comment! {}", comment)).into());
        }
    }

    Ok(n.clone())
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Copy, Clone)]
#[serde(rename_all = "lowercase")]
pub enum NetworkFilterChainProtocol {
    Root,
    Mac,
    STP,
    VLAN,
    ARP,
    RARP,
    IPv4,
    IPv6,
}

impl NetworkFilterChainProtocol {
    pub fn from_xml(xml: &str) -> anyhow::Result<Self> {
        Ok(match xml {
            "root" => Self::Root,
            "mac" => Self::Mac,
            "stp" => Self::STP,
            "vlan" => Self::VLAN,
            "arp" => Self::ARP,
            "rarp" => Self::RARP,
            "ipv4" => Self::IPv4,
            "ipv6" => Self::IPv6,
            _ => {
                return Err(LibVirtStructError::ParseFilteringChain(format!(
                    "Unknown filtering chain: {xml}! "
                ))
                .into());
            }
        })
    }

    pub fn to_xml(&self) -> String {
        match self {
            Self::Root => "root",
            Self::Mac => "mac",
            Self::STP => "stp",
            Self::VLAN => "vlan",
            Self::ARP => "arp",
            Self::RARP => "rarp",
            Self::IPv4 => "ipv4",
            Self::IPv6 => "ipv6",
        }
        .to_string()
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilterChain {
    protocol: NetworkFilterChainProtocol,
    suffix: Option<String>,
}

impl NetworkFilterChain {
    pub fn from_xml(xml: &str) -> anyhow::Result<Self> {
        Ok(match xml.split_once('-') {
            None => Self {
                protocol: NetworkFilterChainProtocol::from_xml(xml)?,
                suffix: None,
            },
            Some((prefix, suffix)) => Self {
                protocol: NetworkFilterChainProtocol::from_xml(prefix)?,
                suffix: Some(suffix.to_string()),
            },
        })
    }

    pub fn to_xml(&self) -> String {
        match &self.suffix {
            None => self.protocol.to_xml(),
            Some(s) => format!("{}-{s}", self.protocol.to_xml()),
        }
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Copy, Clone)]
#[serde(rename_all = "lowercase")]
pub enum NetworkFilterAction {
    /// matching the rule silently discards the packet with no further analysis
    Drop,
    /// matching the rule generates an ICMP reject message with no further analysis
    Reject,
    /// matching the rule accepts the packet with no further analysis
    Accept,
    /// matching the rule passes this filter, but returns control to the calling filter for further
    /// analysis
    Return,
    /// matching the rule goes on to the next rule for further analysis
    Continue,
}

impl NetworkFilterAction {
    pub fn from_xml(xml: &str) -> anyhow::Result<Self> {
        Ok(match xml {
            "drop" => Self::Drop,
            "reject" => Self::Reject,
            "accept" => Self::Accept,
            "return" => Self::Return,
            "continue" => Self::Continue,
            s => {
                return Err(LibVirtStructError::ParseFilteringChain(format!(
                    "Unkown filter action {s}!"
                ))
                .into());
            }
        })
    }

    pub fn to_xml(&self) -> String {
        match self {
            Self::Drop => "drop",
            Self::Reject => "reject",
            Self::Accept => "accept",
            Self::Return => "return",
            Self::Continue => "continue",
        }
        .to_string()
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
#[serde(rename_all = "lowercase")]
pub enum NetworkFilterDirection {
    In,
    Out,
    InOut,
}

impl NetworkFilterDirection {
    pub fn from_xml(xml: &str) -> anyhow::Result<Self> {
        Ok(match xml {
            "in" => Self::In,
            "out" => Self::Out,
            "inout" => Self::InOut,
            s => {
                return Err(LibVirtStructError::ParseFilteringChain(format!(
                    "Unkown filter direction {s}!"
                ))
                .into());
            }
        })
    }

    pub fn to_xml(&self) -> String {
        match self {
            NetworkFilterDirection::In => "in",
            NetworkFilterDirection::Out => "out",
            NetworkFilterDirection::InOut => "inout",
        }
        .to_string()
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Copy, Clone)]
pub enum Layer4State {
    NEW,
    ESTABLISHED,
    RELATED,
    INVALID,
    NONE,
}

impl Layer4State {
    pub fn from_xml(xml: &str) -> anyhow::Result<Self> {
        Ok(match xml {
            "NEW" => Self::NEW,
            "ESTABLISHED" => Self::ESTABLISHED,
            "RELATED" => Self::RELATED,
            "INVALID" => Self::INVALID,
            "NONE" => Self::NONE,
            s => {
                return Err(LibVirtStructError::ParseFilteringChain(format!(
                    "Unkown layer4 state '{s}'!"
                ))
                .into());
            }
        })
    }

    pub fn to_xml(&self) -> String {
        match self {
            Self::NEW => "NEW",
            Self::ESTABLISHED => "ESTABLISHED",
            Self::RELATED => "RELATED",
            Self::INVALID => "INVALID",
            Self::NONE => "NONE",
        }
        .to_string()
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkSelectorMac {
    src_mac_addr: Option<NetworkFilterMacAddressOrVar>,
    src_mac_mask: Option<NetworkFilterMacAddressOrVar>,
    dst_mac_addr: Option<NetworkFilterMacAddressOrVar>,
    dst_mac_mask: Option<NetworkFilterMacAddressOrVar>,
    comment: Option<String>,
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkSelectorARP {
    srcmacaddr: Option<NetworkFilterMacAddressOrVar>,
    srcmacmask: Option<NetworkFilterMacAddressOrVar>,
    dstmacaddr: Option<NetworkFilterMacAddressOrVar>,
    dstmacmask: Option<NetworkFilterMacAddressOrVar>,
    arpsrcipaddr: Option<NetworkFilterIPv4OrVar>,
    arpsrcipmask: Option<u8>,
    arpdstipaddr: Option<NetworkFilterIPv4OrVar>,
    arpdstipmask: Option<u8>,
    comment: Option<String>,
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilterSelectorIP<const V: usize> {
    srcmacaddr: Option<NetworkFilterMacAddressOrVar>,
    srcmacmask: Option<NetworkFilterMacAddressOrVar>,
    dstmacaddr: Option<NetworkFilterMacAddressOrVar>,
    dstmacmask: Option<NetworkFilterMacAddressOrVar>,
    srcipaddr: Option<NetworkFilterIPOrVar<V>>,
    srcipmask: Option<u8>,
    dstipaddr: Option<NetworkFilterIPOrVar<V>>,
    dstipmask: Option<u8>,
    comment: Option<String>,
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilterSelectorLayer4<IPv> {
    srcmacaddr: Option<NetworkFilterMacAddressOrVar>,
    srcipaddr: Option<IPv>,
    srcipmask: Option<u8>,
    dstipaddr: Option<IPv>,
    dstipmask: Option<u8>,
    /// Start of range of source IP address
    srcipfrom: Option<IPv>,
    /// End of range of source IP address
    srcipto: Option<IPv>,
    /// Start of range of destination IP address
    dstipfrom: Option<IPv>,
    /// End of range of destination IP address
    dstipto: Option<IPv>,
    srcportstart: Option<u16>,
    srcportend: Option<u16>,
    dstportstart: Option<u16>,
    dstportend: Option<u16>,
    state: Option<Layer4State>,
    comment: Option<String>,
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkSelectorAll<IPv> {
    comment: Option<String>,
    srcmacaddr: Option<NetworkFilterMacAddressOrVar>,
    srcipaddr: Option<IPv>,
    srcipmask: Option<u8>,
    dstipaddr: Option<IPv>,
    dstipmask: Option<u8>,
    /// Start of range of source IP address
    srcipfrom: Option<IPv>,
    /// End of range of source IP address
    srcipto: Option<IPv>,
    /// Start of range of destination IP address
    dstipfrom: Option<IPv>,
    /// End of range of destination IP address
    dstipto: Option<IPv>,
    state: Option<Layer4State>,
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
#[serde(tag = "type", rename_all = "lowercase")]
pub enum NetworkFilterSelector {
    Mac(NetworkSelectorMac),
    Arp(NetworkSelectorARP),
    Rarp(NetworkSelectorARP),
    IPv4(NetworkFilterSelectorIP<4>),
    IPv6(NetworkFilterSelectorIP<6>),
    TCP(NetworkFilterSelectorLayer4<Ipv4Addr>),
    UDP(NetworkFilterSelectorLayer4<Ipv4Addr>),
    SCTP(NetworkFilterSelectorLayer4<Ipv4Addr>),
    ICMP(NetworkFilterSelectorLayer4<Ipv4Addr>),
    All(NetworkSelectorAll<Ipv4Addr>),
    TCPipv6(NetworkFilterSelectorLayer4<Ipv6Addr>),
    UDPipv6(NetworkFilterSelectorLayer4<Ipv6Addr>),
    SCTPipv6(NetworkFilterSelectorLayer4<Ipv6Addr>),
    ICMPipv6(NetworkFilterSelectorLayer4<Ipv6Addr>),
    Allipv6(NetworkSelectorAll<Ipv6Addr>),
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilterRule {
    action: NetworkFilterAction,
    direction: NetworkFilterDirection,
    /// optional; the priority of the rule controls the order in which the rule will be instantiated
    /// relative to other rules
    ///
    /// Valid values are in the range of -1000 to 1000.
    priority: Option<i32>,
    selectors: Vec<NetworkFilterSelector>,
}

/// Network filter definition
#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
pub struct NetworkFilter {
    pub name: NetworkFilterName,
    chain: Option<NetworkFilterChain>,
    priority: Option<i32>,
    pub uuid: Option<XMLUuid>,
    /// Referenced filters rules
    join_filters: Vec<NetworkFilterName>,
    rules: Vec<NetworkFilterRule>,
}

impl NetworkFilter {
    fn lib2rest_process_mac_rule(n: &NetworkFilterRuleProtocolMac) -> NetworkFilterSelector {
        NetworkFilterSelector::Mac(NetworkSelectorMac {
            src_mac_addr: n.srcmacaddr.as_ref().map(|v| v.into()),
            src_mac_mask: n.srcmacmask.as_ref().map(|v| v.into()),
            dst_mac_addr: n.dstmacaddr.as_ref().map(|v| v.into()),
            dst_mac_mask: n.dstmacmask.as_ref().map(|v| v.into()),
            comment: n.comment.clone(),
        })
    }

    fn lib2rest_process_arp_rule(n: &NetworkFilterRuleProtocolArpXML) -> NetworkSelectorARP {
        NetworkSelectorARP {
            srcmacaddr: n.srcmacaddr.as_ref().map(|v| v.into()),
            srcmacmask: n.srcmacmask.as_ref().map(|v| v.into()),
            dstmacaddr: n.dstmacaddr.as_ref().map(|v| v.into()),
            dstmacmask: n.dstmacmask.as_ref().map(|v| v.into()),
            arpsrcipaddr: n.arpsrcipaddr.as_ref().map(|v| v.into()),
            arpsrcipmask: n.arpsrcipmask,
            arpdstipaddr: n.arpdstipaddr.as_ref().map(|v| v.into()),
            arpdstipmask: n.arpdstipmask,
            comment: n.comment.clone(),
        }
    }

    fn lib2rest_process_ip_rule<const V: usize>(
        n: &NetworkFilterRuleProtocolIpvx,
    ) -> NetworkFilterSelectorIP<V> {
        NetworkFilterSelectorIP {
            srcmacaddr: n.srcmacaddr.as_ref().map(|v| v.into()),
            srcmacmask: n.srcmacmask.as_ref().map(|v| v.into()),
            dstmacaddr: n.dstmacaddr.as_ref().map(|v| v.into()),
            dstmacmask: n.dstmacmask.as_ref().map(|v| v.into()),
            srcipaddr: n.srcipaddr.as_ref().map(|v| v.into()),
            srcipmask: n.srcipmask,
            dstipaddr: n.dstipaddr.as_ref().map(|v| v.into()),
            dstipmask: n.dstipmask,
            comment: n.comment.clone(),
        }
    }

    fn lib2rest_process_layer4_rule<IPv: Copy>(
        n: &NetworkFilterRuleProtocolLayer4<IPv>,
    ) -> anyhow::Result<NetworkFilterSelectorLayer4<IPv>> {
        Ok(NetworkFilterSelectorLayer4 {
            srcmacaddr: n.srcmacaddr.as_ref().map(|v| v.into()),
            srcipaddr: n.srcipaddr,
            srcipmask: n.srcipmask,
            dstipaddr: n.dstipaddr,
            dstipmask: n.dstipmask,
            srcipfrom: n.srcipfrom,
            srcipto: n.srcipto,
            dstipfrom: n.dstipfrom,
            dstipto: n.dstipto,
            srcportstart: n.srcportstart,
            srcportend: n.srcportend,
            dstportstart: n.dstportstart,
            dstportend: n.dstportend,
            state: n.state.as_deref().map(Layer4State::from_xml).transpose()?,
            comment: n.comment.clone(),
        })
    }

    fn lib2rest_process_all_rule<IPv: Copy>(
        n: &NetworkFilterRuleProtocolAllXML<IPv>,
    ) -> anyhow::Result<NetworkSelectorAll<IPv>> {
        Ok(NetworkSelectorAll {
            srcmacaddr: n.srcmacaddr.as_ref().map(|v| v.into()),
            srcipaddr: n.srcipaddr,
            srcipmask: n.srcipmask,
            dstipaddr: n.dstipaddr,
            dstipmask: n.dstipmask,
            srcipfrom: n.srcipfrom,
            srcipto: n.srcipto,
            dstipfrom: n.dstipfrom,
            dstipto: n.dstipto,
            state: n.state.as_deref().map(Layer4State::from_xml).transpose()?,
            comment: n.comment.clone(),
        })
    }

    pub fn lib2rest(xml: NetworkFilterXML) -> anyhow::Result<Self> {
        let mut rules = Vec::with_capacity(xml.rules.len());
        for rule in &xml.rules {
            let mut selectors = Vec::new();

            // Mac selectors
            selectors.append(
                &mut rule
                    .mac_selectors
                    .iter()
                    .map(Self::lib2rest_process_mac_rule)
                    .collect(),
            );

            // ARP - RARP selectors
            selectors.append(
                &mut rule
                    .arp_selectors
                    .iter()
                    .map(|r| NetworkFilterSelector::Arp(Self::lib2rest_process_arp_rule(r)))
                    .collect(),
            );
            selectors.append(
                &mut rule
                    .rarp_selectors
                    .iter()
                    .map(|r| NetworkFilterSelector::Rarp(Self::lib2rest_process_arp_rule(r)))
                    .collect(),
            );

            // IPv4 - IPv6 selectors
            selectors.append(
                &mut rule
                    .ipv4_selectors
                    .iter()
                    .map(|r| NetworkFilterSelector::IPv4(Self::lib2rest_process_ip_rule(r)))
                    .collect(),
            );
            selectors.append(
                &mut rule
                    .ipv6_selectors
                    .iter()
                    .map(|r| NetworkFilterSelector::IPv6(Self::lib2rest_process_ip_rule(r)))
                    .collect(),
            );

            // Layer 4 protocols selectors
            selectors.append(
                &mut rule
                    .tcp_selectors
                    .iter()
                    .map(|r| {
                        Ok(NetworkFilterSelector::TCP(
                            Self::lib2rest_process_layer4_rule(r)?,
                        ))
                    })
                    .collect::<Result<Vec<_>, anyhow::Error>>()?,
            );
            selectors.append(
                &mut rule
                    .udp_selectors
                    .iter()
                    .map(|r| {
                        Ok(NetworkFilterSelector::UDP(
                            Self::lib2rest_process_layer4_rule(r)?,
                        ))
                    })
                    .collect::<Result<Vec<_>, anyhow::Error>>()?,
            );
            selectors.append(
                &mut rule
                    .sctp_selectors
                    .iter()
                    .map(|r| {
                        Ok(NetworkFilterSelector::SCTP(
                            Self::lib2rest_process_layer4_rule(r)?,
                        ))
                    })
                    .collect::<Result<Vec<_>, anyhow::Error>>()?,
            );
            selectors.append(
                &mut rule
                    .icmp_selectors
                    .iter()
                    .map(|r| {
                        Ok(NetworkFilterSelector::ICMP(
                            Self::lib2rest_process_layer4_rule(r)?,
                        ))
                    })
                    .collect::<Result<Vec<_>, anyhow::Error>>()?,
            );

            selectors.append(
                &mut rule
                    .tcp_ipv6_selectors
                    .iter()
                    .map(|r| {
                        Ok(NetworkFilterSelector::TCPipv6(
                            Self::lib2rest_process_layer4_rule(r)?,
                        ))
                    })
                    .collect::<Result<Vec<_>, anyhow::Error>>()?,
            );
            selectors.append(
                &mut rule
                    .udp_ipv6_selectors
                    .iter()
                    .map(|r| {
                        Ok(NetworkFilterSelector::UDPipv6(
                            Self::lib2rest_process_layer4_rule(r)?,
                        ))
                    })
                    .collect::<Result<Vec<_>, anyhow::Error>>()?,
            );
            selectors.append(
                &mut rule
                    .sctp_ipv6_selectors
                    .iter()
                    .map(|r| {
                        Ok(NetworkFilterSelector::SCTPipv6(
                            Self::lib2rest_process_layer4_rule(r)?,
                        ))
                    })
                    .collect::<Result<Vec<_>, anyhow::Error>>()?,
            );
            selectors.append(
                &mut rule
                    .imcp_ipv6_selectors
                    .iter()
                    .map(|r| {
                        Ok(NetworkFilterSelector::ICMPipv6(
                            Self::lib2rest_process_layer4_rule(r)?,
                        ))
                    })
                    .collect::<Result<Vec<_>, anyhow::Error>>()?,
            );

            // All selectors
            selectors.append(
                &mut rule
                    .all_selectors
                    .iter()
                    .map(|r| {
                        Ok(NetworkFilterSelector::All(Self::lib2rest_process_all_rule(
                            r,
                        )?))
                    })
                    .collect::<Result<Vec<_>, anyhow::Error>>()?,
            );

            selectors.append(
                &mut rule
                    .all_ipv6_selectors
                    .iter()
                    .map(|r| {
                        Ok(NetworkFilterSelector::Allipv6(
                            Self::lib2rest_process_all_rule(r)?,
                        ))
                    })
                    .collect::<Result<Vec<_>, anyhow::Error>>()?,
            );

            rules.push(NetworkFilterRule {
                action: NetworkFilterAction::from_xml(&rule.action)?,
                direction: NetworkFilterDirection::from_xml(&rule.direction)?,
                priority: rule.priority,
                selectors,
            })
        }

        Ok(Self {
            name: NetworkFilterName(xml.name),
            uuid: xml.uuid,
            chain: xml
                .chain
                .as_deref()
                .map(NetworkFilterChain::from_xml)
                .transpose()?,
            priority: xml.priority,
            join_filters: xml
                .filterrefs
                .iter()
                .map(|i| NetworkFilterName(i.filter.to_string()))
                .collect(),
            rules,
        })
    }

    fn rest2lib_process_arp_selector(
        selector: &NetworkSelectorARP,
    ) -> anyhow::Result<NetworkFilterRuleProtocolArpXML> {
        Ok(NetworkFilterRuleProtocolArpXML {
            srcmacaddr: extract_mac_address_or_var(&selector.srcmacaddr)?,
            srcmacmask: extract_mac_address_or_var(&selector.srcmacmask)?,
            dstmacaddr: extract_mac_address_or_var(&selector.dstmacaddr)?,
            dstmacmask: extract_mac_address_or_var(&selector.dstmacmask)?,
            arpsrcipaddr: extract_ip_or_var(&selector.arpsrcipaddr)?,
            arpsrcipmask: selector.arpsrcipmask,
            arpdstipaddr: extract_ip_or_var(&selector.arpdstipaddr)?,
            arpdstipmask: selector.arpdstipmask,
            comment: extract_nw_filter_comment(&selector.comment)?,
        })
    }

    fn rest2lib_process_ip_selector<const V: usize>(
        selector: &NetworkFilterSelectorIP<V>,
    ) -> anyhow::Result<NetworkFilterRuleProtocolIpvx> {
        Ok(NetworkFilterRuleProtocolIpvx {
            srcmacaddr: extract_mac_address_or_var(&selector.srcmacaddr)?,
            srcmacmask: extract_mac_address_or_var(&selector.srcmacmask)?,
            dstmacaddr: extract_mac_address_or_var(&selector.dstmacaddr)?,
            dstmacmask: extract_mac_address_or_var(&selector.dstmacmask)?,
            srcipaddr: extract_ip_or_var(&selector.srcipaddr)?,
            srcipmask: extract_ip_mask::<V>(selector.srcipmask)?,
            dstipaddr: extract_ip_or_var(&selector.dstipaddr)?,
            dstipmask: extract_ip_mask::<V>(selector.dstipmask)?,
            comment: extract_nw_filter_comment(&selector.comment)?,
        })
    }

    fn rest2lib_process_layer4_selector<IPv: Copy>(
        selector: &NetworkFilterSelectorLayer4<IPv>,
    ) -> anyhow::Result<NetworkFilterRuleProtocolLayer4<IPv>> {
        Ok(NetworkFilterRuleProtocolLayer4 {
            srcmacaddr: extract_mac_address_or_var(&selector.srcmacaddr)?,
            srcipaddr: selector.srcipaddr,
            // This IP mask is not checked
            srcipmask: selector.srcipmask,
            dstipaddr: selector.dstipaddr,
            // This IP mask is not checked
            dstipmask: selector.dstipmask,
            srcipfrom: selector.srcipfrom,
            srcipto: selector.srcipto,
            dstipfrom: selector.dstipfrom,
            dstipto: selector.dstipto,
            srcportstart: selector.srcportstart,
            srcportend: selector.srcportend,
            dstportstart: selector.dstportstart,
            dstportend: selector.dstportend,
            state: selector.state.map(|s| s.to_xml()),
            comment: extract_nw_filter_comment(&selector.comment)?,
        })
    }

    fn rest2lib_process_all_selector<IPv: Copy>(
        selector: &NetworkSelectorAll<IPv>,
    ) -> anyhow::Result<NetworkFilterRuleProtocolAllXML<IPv>> {
        Ok(NetworkFilterRuleProtocolAllXML {
            srcmacaddr: extract_mac_address_or_var(&selector.srcmacaddr)?,
            srcipaddr: selector.srcipaddr,
            // This IP mask is not checked
            srcipmask: selector.srcipmask,
            dstipaddr: selector.dstipaddr,
            // This IP mask is not checked
            dstipmask: selector.dstipmask,
            srcipfrom: selector.srcipfrom,
            srcipto: selector.srcipto,
            dstipfrom: selector.dstipfrom,
            dstipto: selector.dstipto,
            state: selector.state.map(|s| s.to_xml()),
            comment: extract_nw_filter_comment(&selector.comment)?,
        })
    }

    fn rest2lib_process_rule(rule: &NetworkFilterRule) -> anyhow::Result<NetworkFilterRuleXML> {
        let mut rule_xml = NetworkFilterRuleXML {
            action: rule.action.to_xml(),
            direction: rule.direction.to_xml(),
            priority: rule.priority,
            ..Default::default()
        };

        for sel in &rule.selectors {
            match sel {
                NetworkFilterSelector::Mac(mac) => {
                    rule_xml.mac_selectors.push(NetworkFilterRuleProtocolMac {
                        srcmacaddr: extract_mac_address_or_var(&mac.src_mac_addr)?,
                        srcmacmask: extract_mac_address_or_var(&mac.src_mac_mask)?,
                        dstmacaddr: extract_mac_address_or_var(&mac.dst_mac_addr)?,
                        dstmacmask: extract_mac_address_or_var(&mac.dst_mac_mask)?,
                        comment: extract_nw_filter_comment(&mac.comment)?,
                    })
                }

                NetworkFilterSelector::Arp(a) => {
                    rule_xml
                        .arp_selectors
                        .push(Self::rest2lib_process_arp_selector(a)?);
                }

                NetworkFilterSelector::Rarp(a) => {
                    rule_xml
                        .rarp_selectors
                        .push(Self::rest2lib_process_arp_selector(a)?);
                }

                NetworkFilterSelector::IPv4(ip) => rule_xml
                    .ipv4_selectors
                    .push(Self::rest2lib_process_ip_selector(ip)?),
                NetworkFilterSelector::IPv6(ip) => rule_xml
                    .ipv6_selectors
                    .push(Self::rest2lib_process_ip_selector(ip)?),

                NetworkFilterSelector::TCP(tcp) => rule_xml
                    .tcp_selectors
                    .push(Self::rest2lib_process_layer4_selector(tcp)?),

                NetworkFilterSelector::UDP(udp) => rule_xml
                    .udp_selectors
                    .push(Self::rest2lib_process_layer4_selector(udp)?),

                NetworkFilterSelector::SCTP(sctp) => rule_xml
                    .sctp_selectors
                    .push(Self::rest2lib_process_layer4_selector(sctp)?),

                NetworkFilterSelector::ICMP(icmp) => rule_xml
                    .icmp_selectors
                    .push(Self::rest2lib_process_layer4_selector(icmp)?),

                NetworkFilterSelector::All(all) => {
                    rule_xml
                        .all_selectors
                        .push(Self::rest2lib_process_all_selector(all)?);
                }

                NetworkFilterSelector::TCPipv6(tcpv6) => rule_xml
                    .tcp_ipv6_selectors
                    .push(Self::rest2lib_process_layer4_selector(tcpv6)?),

                NetworkFilterSelector::UDPipv6(udpv6) => rule_xml
                    .udp_ipv6_selectors
                    .push(Self::rest2lib_process_layer4_selector(udpv6)?),

                NetworkFilterSelector::SCTPipv6(sctpv6) => rule_xml
                    .sctp_ipv6_selectors
                    .push(Self::rest2lib_process_layer4_selector(sctpv6)?),

                NetworkFilterSelector::ICMPipv6(icmpv6) => rule_xml
                    .imcp_ipv6_selectors
                    .push(Self::rest2lib_process_layer4_selector(icmpv6)?),

                NetworkFilterSelector::Allipv6(all) => {
                    rule_xml
                        .all_ipv6_selectors
                        .push(Self::rest2lib_process_all_selector(all)?);
                }
            }
        }

        Ok(rule_xml)
    }

    pub fn rest2lib(&self) -> anyhow::Result<NetworkFilterXML> {
        if !self.name.is_valid() {
            return Err(
                NetworkFilterExtraction("Network filter name is invalid!".to_string()).into(),
            );
        }

        if let Some(priority) = self.priority {
            if !(-1000..=1000).contains(&priority) {
                return Err(
                    NetworkFilterExtraction("Network priority is invalid!".to_string()).into(),
                );
            }
        }

        for fref in &self.join_filters {
            if !fref.is_valid() {
                return Err(
                    StructureExtraction("Referenced network filter name is invalid!").into(),
                );
            }
        }

        let mut rules = Vec::with_capacity(self.rules.len());

        for rule in &self.rules {
            rules.push(Self::rest2lib_process_rule(rule)?);
        }

        Ok(NetworkFilterXML {
            name: self.name.0.to_string(),
            uuid: self.uuid,
            chain: self.chain.as_ref().map(|c| c.to_xml()),
            priority: self.priority,
            filterrefs: self
                .join_filters
                .iter()
                .map(|jf| NetworkFilterRefXML {
                    filter: jf.0.to_string(),
                })
                .collect::<Vec<_>>(),
            rules,
        })
    }
}

#[cfg(test)]
mod test {
    use crate::libvirt_rest_structures::nw_filter::is_var_def;

    #[test]
    pub fn var_def() {
        assert!(is_var_def("$MAC"));
        assert!(is_var_def("$MAC_ADDRESS"));

        assert!(!is_var_def("$$MAC"));
        assert!(!is_var_def("$$MACé"));
        assert!(!is_var_def("$$MAC@"));
        assert!(!is_var_def("$$MAC TEST"));
    }
}
