import { Button } from "@mui/material";
import React from "react";
import { useNavigate, useParams } from "react-router-dom";
import { useAlert } from "../hooks/providers/AlertDialogProvider";
import { useLoadingMessage } from "../hooks/providers/LoadingMessageProvider";
import { useSnackbar } from "../hooks/providers/SnackbarProvider";
import { AsyncWidget } from "../widgets/AsyncWidget";
import { ConfigImportExportButtons } from "../widgets/ConfigImportExportButtons";
import { VirtWebRouteContainer } from "../widgets/VirtWebRouteContainer";
import { NWFilter, NWFilterApi, NWFilterURL } from "../api/NWFilterApi";
import { NWFilterDetails } from "../widgets/nwfilter/NWFilterDetails";

export function CreateNWFilterRoute(): React.ReactElement {
  const alert = useAlert();
  const snackbar = useSnackbar();
  const navigate = useNavigate();

  const [nwfilter, setNWFilter] = React.useState<NWFilter>({
    name: "my-filter",
    chain: { protocol: "root" },
    join_filters: [],
    rules: [],
  });

  const createNWFilter = async (n: NWFilter) => {
    try {
      const res = await NWFilterApi.Create(n);
      snackbar("The network filter was successfully created!");
      navigate(`/nwfilter/${res.uid}`);
    } catch (e) {
      console.error(e);
      alert(`Failed to create network filter!\n${e}`);
    }
  };

  return (
    <EditNetworkFilterRouteInner
      nwfilter={nwfilter}
      creating={true}
      onCancel={() => navigate("/nwfilter")}
      onSave={createNWFilter}
      onReplace={setNWFilter}
    />
  );
}

export function EditNWFilterRoute(): React.ReactElement {
  const alert = useAlert();
  const snackbar = useSnackbar();

  const { uuid } = useParams();
  const navigate = useNavigate();

  const [nwfilter, setNWFilter] = React.useState<NWFilter | undefined>();

  const load = async () => {
    setNWFilter(await NWFilterApi.GetSingle(uuid!));
  };

  const updateNetworkFilter = async (n: NWFilter) => {
    try {
      await NWFilterApi.Update(n);
      snackbar("The network filter was successfully updated!");
      navigate(NWFilterURL(nwfilter!));
    } catch (e) {
      console.error(e);
      alert(`Failed to update network filter!\n${e}`);
    }
  };

  return (
    <AsyncWidget
      loadKey={uuid}
      ready={nwfilter !== undefined}
      errMsg="Failed to fetch network filter information!"
      load={load}
      build={() => (
        <EditNetworkFilterRouteInner
          nwfilter={nwfilter!}
          creating={false}
          onCancel={() => navigate(`/nwfilter/${uuid}`)}
          onSave={updateNetworkFilter}
          onReplace={setNWFilter}
        />
      )}
    />
  );
}

function EditNetworkFilterRouteInner(p: {
  nwfilter: NWFilter;
  creating: boolean;
  onCancel: () => void;
  onSave: (vm: NWFilter) => Promise<void>;
  onReplace: (vm: NWFilter) => void;
}): React.ReactElement {
  const loadingMessage = useLoadingMessage();

  const [changed, setChanged] = React.useState(false);

  const [, updateState] = React.useState<any>();
  const forceUpdate = React.useCallback(() => { updateState({}); }, []);

  const valueChanged = () => {
    setChanged(true);
    forceUpdate();
  };

  const save = async () => {
    loadingMessage.show("Saving network filter configuration...");
    await p.onSave(p.nwfilter);
    loadingMessage.hide();
  };

  return (
    <VirtWebRouteContainer
      label={p.creating ? "Create a Network Filter" : "Edit Network Filter"}
      actions={
        <span>
          <ConfigImportExportButtons
            currentConf={p.nwfilter}
            filename={`nwfilter-${p.nwfilter.name}.json`}
            importConf={(c) => {
              p.onReplace(c);
              valueChanged();
            }}
          />

          {changed && (
            <Button
              variant="contained"
              onClick={save}
              style={{ marginRight: "10px" }}
            >
              {p.creating ? "Create" : "Save"}
            </Button>
          )}
          <Button onClick={p.onCancel} variant="outlined">
            Cancel
          </Button>
        </span>
      }
    >
      <NWFilterDetails
        nwfilter={p.nwfilter}
        editable={true}
        onChange={valueChanged}
      />
    </VirtWebRouteContainer>
  );
}
