/* eslint-disable react-x/no-array-index-key */
import ArrowDownwardIcon from "@mui/icons-material/ArrowDownward";
import ArrowUpwardIcon from "@mui/icons-material/ArrowUpward";
import DeleteIcon from "@mui/icons-material/Delete";
import PlaylistAddIcon from "@mui/icons-material/PlaylistAdd";
import {
  Button,
  Card,
  CardActions,
  CardContent,
  IconButton,
  Paper,
  Tooltip,
} from "@mui/material";
import {
  NWFSAllBase,
  NWFSArpOrRARP,
  NWFSIPBase,
  NWFSLayer4Base,
  NWFSMac,
  NWFSelector,
  NWFilterRule,
} from "../../api/NWFilterApi";
import { ServerApi } from "../../api/ServerApi";
import { EditSection } from "./EditSection";
import { IPInput, IPInputWithMask } from "./IPInput";
import { MACInput } from "./MACInput";
import { NWFConnStateInput } from "./NWFConnStateInput";
import { NWFilterPriorityInput } from "./NWFilterPriorityInput";
import { PortInput } from "./PortInput";
import { SelectInput } from "./SelectInput";
import { TextInput } from "./TextInput";

export function NWFilterRules(p: {
  editable: boolean;
  rules: NWFilterRule[];
  onChange?: () => void;
}): React.ReactElement {
  const addRule = () => {
    p.rules.push({
      action: "drop",
      direction: "inout",
      selectors: [],
    });
    p.onChange?.();
  };

  const swapRules = (f: number, s: number) => {
    const swap = p.rules[f];
    p.rules[f] = p.rules[s];
    p.rules[s] = swap;
    p.onChange?.();
  };

  const deleteRule = (num: number) => {
    p.rules.splice(num, 1);
    p.onChange?.();
  };

  return (
    <EditSection title="Rules">
      {p.rules.map((r, n) => (
        <NWRuleEdit
          key={n}
          rule={r}
          onDelete={() => {
            deleteRule(n);
          }}
          onGoDown={
            n < p.rules.length - 1
              ? () => {
                  swapRules(n, n + 1);
                }
              : undefined
          }
          onGoUp={
            n > 0
              ? () => {
                  swapRules(n, n - 1);
                }
              : undefined
          }
          {...p}
        />
      ))}

      <div style={{ textAlign: "right" }}>
        {p.editable && <Button onClick={addRule}>Add a new rule</Button>}
      </div>
    </EditSection>
  );
}

function NWRuleEdit(p: {
  editable: boolean;
  rule: NWFilterRule;
  onChange?: () => void;
  onGoUp?: () => void;
  onGoDown?: () => void;
  onDelete: () => void;
}): React.ReactElement {
  const addSelector = () => {
    p.rule.selectors.push({
      type: "all",
    });
    p.onChange?.();
  };

  const deleteSelector = (num: number) => {
    p.rule.selectors.splice(num, 1);
    p.onChange?.();
  };

  return (
    <Card style={{ margin: "30px" }} elevation={3}>
      <CardContent>
        <div style={{ display: "flex" }}>
          <SelectInput
            editable={p.editable}
            label="Action"
            value={p.rule.action}
            onValueChange={(v) => {
              p.rule.action = v as any;
              p.onChange?.();
            }}
            options={[
              { label: "drop", value: "drop" },
              { label: "reject", value: "reject" },
              { label: "accept", value: "accept" },
              { label: "return", value: "return" },
              { label: "continue", value: "continue" },
            ]}
          />
          <span style={{ width: "20px" }}></span>
          <SelectInput
            editable={p.editable}
            label="Direction"
            value={p.rule.direction}
            onValueChange={(v) => {
              p.rule.direction = v as any;
              p.onChange?.();
            }}
            options={[
              { label: "in", value: "in" },
              { label: "out", value: "out" },
              { label: "inout", value: "inout" },
            ]}
          />
          <span style={{ width: "20px" }}></span>
          <NWFilterPriorityInput
            {...p}
            label="Priority"
            value={p.rule.priority}
            onChange={(v) => {
              p.rule.priority = v;
              p.onChange?.();
            }}
          />
        </div>

        {p.rule.selectors.map((s, n) => (
          <NWFSelectorEdit
            key={n}
            editable={p.editable}
            onChange={p.onChange}
            selector={s}
            onDelete={() => {
              deleteSelector(n);
            }}
          />
        ))}
      </CardContent>
      <CardActions>
        {p.editable && (
          <div style={{ display: "flex", width: "100%" }}>
            <Tooltip title="Remove the rule">
              <IconButton color="error" onClick={p.onDelete}>
                <DeleteIcon />
              </IconButton>
            </Tooltip>

            <span style={{ flex: 1 }}></span>

            {ServerApi.Config.constraints.nwfilter_selectors_count.max >
              p.rule.selectors.length && (
              <Tooltip title="Add a selector">
                <IconButton onClick={addSelector}>
                  <PlaylistAddIcon />
                </IconButton>
              </Tooltip>
            )}

            {p.onGoUp && (
              <Tooltip title="Move rule upward">
                <IconButton onClick={p.onGoUp}>
                  <ArrowUpwardIcon />
                </IconButton>
              </Tooltip>
            )}

            {p.onGoDown && (
              <Tooltip title="Move rule downward">
                <IconButton onClick={p.onGoDown}>
                  <ArrowDownwardIcon />
                </IconButton>
              </Tooltip>
            )}
          </div>
        )}
      </CardActions>
    </Card>
  );
}

function NWFSelectorEdit(p: {
  editable: boolean;
  selector: NWFSelector;
  onDelete: () => void;
  onChange?: () => void;
}): React.ReactElement {
  return (
    <Paper elevation={10} style={{ padding: "10px" }}>
      <div style={{ display: "flex", width: "100%" }}>
        <div style={{ flex: 1 }}>
          <SelectInput
            editable={p.editable}
            label="Type"
            onValueChange={(v) => {
              p.selector.type = v! as any;
              p.onChange?.();
            }}
            value={p.selector.type}
            options={[
              { label: "MAC (Ethernet)", value: "mac" },

              { label: "ARP", value: "arp" },
              { label: "RARP", value: "rarp" },

              { label: "IPv4", value: "ipv4" },
              { label: "IPv6", value: "ipv6" },

              { label: "TCP over IPv4", value: "tcp" },
              { label: "UDP over IPv4", value: "udp" },
              { label: "SCTP over IPv4", value: "sctp" },
              { label: "ICMPv4", value: "icmp" },

              { label: "All over IPv4", value: "all" },

              { label: "TCP over IPv6", value: "tcpipv6" },
              { label: "UDP over IPv6", value: "udpipv6" },
              { label: "SCTP over IPv6", value: "sctpipv6" },
              { label: "ICMPv6", value: "icmpipv6" },

              { label: "All over IPv6", value: "allipv6" },
            ]}
          />

          {p.selector.type === "mac" && (
            <NWFSelectorMac {...p} selector={p.selector} />
          )}

          {(p.selector.type === "arp" || p.selector.type === "rarp") && (
            <NWFSelectorArp {...p} selector={p.selector} />
          )}

          {p.selector.type === "ipv4" && (
            <NWFSelectorIP {...p} selector={p.selector} version={4} />
          )}

          {p.selector.type === "ipv6" && (
            <NWFSelectorIP {...p} selector={p.selector} version={6} />
          )}

          {(p.selector.type === "tcp" ||
            p.selector.type === "udp" ||
            p.selector.type === "sctp" ||
            p.selector.type === "icmp") && (
            <NWFSelectorLayer4 {...p} selector={p.selector} version={4} />
          )}

          {p.selector.type === "all" && (
            <NWFSelectorAll {...p} selector={p.selector} version={4} />
          )}

          {(p.selector.type === "tcpipv6" ||
            p.selector.type === "udpipv6" ||
            p.selector.type === "sctpipv6" ||
            p.selector.type === "icmpipv6") && (
            <NWFSelectorLayer4 {...p} selector={p.selector} version={6} />
          )}

          {p.selector.type === "allipv6" && (
            <NWFSelectorAll {...p} selector={p.selector} version={6} />
          )}

          <TextInput
            editable={p.editable}
            label="Comment"
            value={p.selector.comment}
            onValueChange={(v) => {
              p.selector.comment = v;
              p.onChange?.();
            }}
            size={ServerApi.Config.constraints.nwfilter_comment_size}
          />
        </div>

        {p.editable && (
          <div style={{ display: "flex", justifyContent: "center" }}>
            <Tooltip title="Remove the selector">
              <IconButton color="error" onClick={p.onDelete}>
                <DeleteIcon />
              </IconButton>
            </Tooltip>
          </div>
        )}
      </div>
    </Paper>
  );
}

interface SpecificSelectorEditor<E> {
  editable: boolean;
  selector: E;
  onChange?: () => void;
}

interface SpecificSelectorEditorWithIPVersion<E>
  extends SpecificSelectorEditor<E> {
  version: 4 | 6;
}

function NWFSelectorMac(
  p: SpecificSelectorEditor<NWFSMac>
): React.ReactElement {
  return (
    <>
      <MACInput
        {...p}
        label="Src mac address"
        value={p.selector.src_mac_addr}
        onValueChange={(v) => {
          p.selector.src_mac_addr = v;
          p.onChange?.();
        }}
      />
      <MACInput
        {...p}
        label="Src mac mask"
        value={p.selector.src_mac_mask}
        onValueChange={(v) => {
          p.selector.src_mac_mask = v;
          p.onChange?.();
        }}
      />
      <MACInput
        {...p}
        label="Dst mac address"
        value={p.selector.dst_mac_addr}
        onValueChange={(v) => {
          p.selector.dst_mac_addr = v;
          p.onChange?.();
        }}
      />
      <MACInput
        {...p}
        label="Dst mac mask"
        value={p.selector.dst_mac_mask}
        onValueChange={(v) => {
          p.selector.dst_mac_mask = v;
          p.onChange?.();
        }}
      />
    </>
  );
}

function NWFSelectorArp(
  p: SpecificSelectorEditor<NWFSArpOrRARP>
): React.ReactElement {
  return (
    <>
      <MACInput
        {...p}
        label="Src mac address"
        value={p.selector.srcmacaddr}
        onValueChange={(v) => {
          p.selector.srcmacaddr = v;
          p.onChange?.();
        }}
      />
      <MACInput
        {...p}
        label="Src mac mask"
        value={p.selector.srcmacmask}
        onValueChange={(v) => {
          p.selector.srcmacmask = v;
          p.onChange?.();
        }}
      />
      <MACInput
        {...p}
        label="Dst mac address"
        value={p.selector.dstmacaddr}
        onValueChange={(v) => {
          p.selector.dstmacaddr = v;
          p.onChange?.();
        }}
      />
      <MACInput
        {...p}
        label="Dst mac mask"
        value={p.selector.dstmacmask}
        onValueChange={(v) => {
          p.selector.dstmacmask = v;
          p.onChange?.();
        }}
      />
      <IPInputWithMask
        {...p}
        label="ARP src ip"
        ip={p.selector.arpsrcipaddr}
        mask={p.selector.arpsrcipmask}
        version={4}
        onValueChange={(ip, mask) => {
          p.selector.arpsrcipaddr = ip;
          p.selector.arpsrcipmask = mask;
          p.onChange?.();
        }}
      />
      <IPInputWithMask
        {...p}
        label="ARP dst ip"
        ip={p.selector.arpdstipaddr}
        mask={p.selector.arpdstipmask}
        version={4}
        onValueChange={(ip, mask) => {
          p.selector.arpdstipaddr = ip;
          p.selector.arpdstipmask = mask;
          p.onChange?.();
        }}
      />
    </>
  );
}

function NWFSelectorIP(
  p: SpecificSelectorEditorWithIPVersion<NWFSIPBase>
): React.ReactElement {
  return (
    <>
      <MACInput
        {...p}
        label="Src mac address"
        value={p.selector.srcmacaddr}
        onValueChange={(v) => {
          p.selector.srcmacaddr = v;
          p.onChange?.();
        }}
      />
      <MACInput
        {...p}
        label="Src mac mask"
        value={p.selector.srcmacmask}
        onValueChange={(v) => {
          p.selector.srcmacmask = v;
          p.onChange?.();
        }}
      />
      <MACInput
        {...p}
        label="Dst mac address"
        value={p.selector.dstmacaddr}
        onValueChange={(v) => {
          p.selector.dstmacaddr = v;
          p.onChange?.();
        }}
      />
      <MACInput
        {...p}
        label="Dst mac mask"
        value={p.selector.dstmacmask}
        onValueChange={(v) => {
          p.selector.dstmacmask = v;
          p.onChange?.();
        }}
      />
      <IPInputWithMask
        {...p}
        label="Source IP address / mask"
        ip={p.selector.srcipaddr}
        mask={p.selector.srcipmask}
        version={p.version}
        onValueChange={(ip, mask) => {
          p.selector.srcipaddr = ip;
          p.selector.srcipmask = mask;
          p.onChange?.();
        }}
      />
      <IPInputWithMask
        {...p}
        label="Destination IP address / mask"
        ip={p.selector.dstipaddr}
        mask={p.selector.dstipmask}
        version={p.version}
        onValueChange={(ip, mask) => {
          p.selector.dstipaddr = ip;
          p.selector.dstipmask = mask;
          p.onChange?.();
        }}
      />
    </>
  );
}

function NWFSelectorLayer4(
  p: SpecificSelectorEditorWithIPVersion<NWFSLayer4Base>
): React.ReactElement {
  return (
    <>
      <MACInput
        {...p}
        label="Src mac address"
        value={p.selector.srcmacaddr}
        onValueChange={(v) => {
          p.selector.srcmacaddr = v;
          p.onChange?.();
        }}
      />
      <IPInputWithMask
        {...p}
        label="Source IP address / mask"
        ip={p.selector.srcipaddr}
        mask={p.selector.srcipmask}
        version={p.version}
        onValueChange={(ip, mask) => {
          p.selector.srcipaddr = ip;
          p.selector.srcipmask = mask;
          p.onChange?.();
        }}
      />
      <IPInputWithMask
        {...p}
        label="Destination IP address / mask"
        ip={p.selector.dstipaddr}
        mask={p.selector.dstipmask}
        version={p.version}
        onValueChange={(ip, mask) => {
          p.selector.dstipaddr = ip;
          p.selector.dstipmask = mask;
          p.onChange?.();
        }}
      />
      <IPInput
        {...p}
        label="Source IP from"
        value={p.selector.srcipfrom}
        onValueChange={(ip) => {
          p.selector.srcipfrom = ip;
          p.onChange?.();
        }}
      />
      <IPInput
        {...p}
        label="Source IP to"
        value={p.selector.srcipto}
        onValueChange={(ip) => {
          p.selector.srcipto = ip;
          p.onChange?.();
        }}
      />
      <IPInput
        {...p}
        label="Destination IP from"
        value={p.selector.dstipfrom}
        onValueChange={(ip) => {
          p.selector.dstipfrom = ip;
          p.onChange?.();
        }}
      />
      <IPInput
        {...p}
        label="Destination IP to"
        value={p.selector.dstipto}
        onValueChange={(ip) => {
          p.selector.dstipto = ip;
          p.onChange?.();
        }}
      />
      <PortInput
        {...p}
        label="Source port start"
        value={p.selector.srcportstart}
        onChange={(port) => {
          p.selector.srcportstart = port;
          p.onChange?.();
        }}
      />
      <PortInput
        {...p}
        label="Source port end"
        value={p.selector.srcportend}
        onChange={(port) => {
          p.selector.srcportend = port;
          p.onChange?.();
        }}
      />
      <PortInput
        {...p}
        label="Destination port start"
        value={p.selector.dstportstart}
        onChange={(port) => {
          p.selector.dstportstart = port;
          p.onChange?.();
        }}
      />
      <PortInput
        {...p}
        label="Destination port end"
        value={p.selector.dstportend}
        onChange={(port) => {
          p.selector.dstportend = port;
          p.onChange?.();
        }}
      />
      <NWFConnStateInput
        {...p}
        value={p.selector.state}
        onChange={(v) => {
          p.selector.state = v;
          p.onChange?.();
        }}
      />
    </>
  );
}

function NWFSelectorAll(
  p: SpecificSelectorEditorWithIPVersion<NWFSAllBase>
): React.ReactElement {
  return (
    <>
      <MACInput
        {...p}
        label="Src mac address"
        value={p.selector.srcmacaddr}
        onValueChange={(v) => {
          p.selector.srcmacaddr = v;
          p.onChange?.();
        }}
      />
      <IPInputWithMask
        {...p}
        label="Source IP address / mask"
        ip={p.selector.srcipaddr}
        mask={p.selector.srcipmask}
        version={p.version}
        onValueChange={(ip, mask) => {
          p.selector.srcipaddr = ip;
          p.selector.srcipmask = mask;
          p.onChange?.();
        }}
      />
      <IPInputWithMask
        {...p}
        label="Destination IP address / mask"
        ip={p.selector.dstipaddr}
        mask={p.selector.dstipmask}
        version={p.version}
        onValueChange={(ip, mask) => {
          p.selector.dstipaddr = ip;
          p.selector.dstipmask = mask;
          p.onChange?.();
        }}
      />
      <IPInput
        {...p}
        label="Source IP from"
        value={p.selector.srcipfrom}
        onValueChange={(ip) => {
          p.selector.srcipfrom = ip;
          p.onChange?.();
        }}
      />
      <IPInput
        {...p}
        label="Source IP to"
        value={p.selector.srcipto}
        onValueChange={(ip) => {
          p.selector.srcipto = ip;
          p.onChange?.();
        }}
      />
      <IPInput
        {...p}
        label="Destination IP from"
        value={p.selector.dstipfrom}
        onValueChange={(ip) => {
          p.selector.dstipfrom = ip;
          p.onChange?.();
        }}
      />
      <IPInput
        {...p}
        label="Destination IP to"
        value={p.selector.dstipto}
        onValueChange={(ip) => {
          p.selector.dstipto = ip;
          p.onChange?.();
        }}
      />
      <NWFConnStateInput
        {...p}
        value={p.selector.state}
        onChange={(v) => {
          p.selector.state = v;
          p.onChange?.();
        }}
      />
    </>
  );
}
