/* eslint-disable react-x/no-array-index-key */
import { mdiNetworkOutline } from "@mdi/js";
import Icon from "@mdi/react";
import DeleteIcon from "@mui/icons-material/Delete";
import {
  Avatar,
  Button,
  IconButton,
  ListItem,
  ListItemAvatar,
  ListItemText,
  Tooltip,
  Typography,
} from "@mui/material";
import Grid from "@mui/material/Grid";
import { NWFilter } from "../../api/NWFilterApi";
import { NetworkInfo } from "../../api/NetworksApi";
import { ServerApi } from "../../api/ServerApi";
import { VMInfo, VMNetInterface } from "../../api/VMApi";
import { useConfirm } from "../../hooks/providers/ConfirmDialogProvider";
import { randomMacAddress } from "../../utils/RandUtils";
import { EditSection } from "./EditSection";
import { MACInput } from "./MACInput";
import { NWFilterSelectInput } from "./NWFilterSelectInput";
import { SelectInput } from "./SelectInput";
import { VMNetworkFilterParameters } from "./VMNetworkFilterParameters";

export function VMNetworksList(p: {
  vm: VMInfo;
  onChange?: () => void;
  editable: boolean;
  networksList: NetworkInfo[];
  bridgesList: string[];
  networkFiltersList: NWFilter[];
}): React.ReactElement {
  const addNew = () => {
    p.vm.networks.push({
      type: "UserspaceSLIRPStack",
      model: "Virtio",
      mac: randomMacAddress(ServerApi.Config.net_mac_prefix),
    });
    p.onChange?.();
  };

  return (
    <>
      {p.editable && (
        <div style={{ textAlign: "right", marginTop: "5px" }}>
          <Button onClick={addNew}>Add a new network interface</Button>
        </div>
      )}

      {p.vm.networks.length === 0 && (
        <Typography style={{ textAlign: "center", paddingTop: "25px" }}>
          No network interface defined yet!
        </Typography>
      )}

      <Grid container spacing={2}>
        {/* networks list */}
        {p.vm.networks.map((n, num) => (
          <EditSection key={num}>
            <NetworkInfoWidget
              network={n}
              removeFromList={() => {
                p.vm.networks.splice(num, 1);
                p.onChange?.();
              }}
              {...p}
            />
          </EditSection>
        ))}
      </Grid>
    </>
  );
}

function NetworkInfoWidget(p: {
  editable: boolean;
  network: VMNetInterface;
  onChange?: () => void;
  removeFromList: () => void;
  networksList: NetworkInfo[];
  bridgesList: string[];
  networkFiltersList: NWFilter[];
}): React.ReactElement {
  const confirm = useConfirm();
  const deleteNetwork = async () => {
    if (
      !(await confirm("Do you really want to remove this network interface?"))
    )
      return;

    p.removeFromList();
    p.onChange?.();
  };

  return (
    <>
      <ListItem
        secondaryAction={
          p.editable && (
            <IconButton
              edge="end"
              aria-label="remove network"
              onClick={deleteNetwork}
            >
              <Tooltip title="Remove network">
                <DeleteIcon />
              </Tooltip>
            </IconButton>
          )
        }
      >
        <ListItemAvatar>
          <Avatar>
            <Icon path={mdiNetworkOutline} />
          </Avatar>
        </ListItemAvatar>
        <ListItemText
          primary={
            p.editable ? (
              <SelectInput
                label=""
                editable
                value={p.network.type}
                onValueChange={(v) => {
                  p.network.type = v as any;
                  p.onChange?.();
                }}
                options={[
                  {
                    label: "Userspace SLIRP stack",
                    value: "UserspaceSLIRPStack",
                    description:
                      "Provides a virtual LAN with NAT to the outside world. The virtual network has DHCP & DNS services",
                  },
                  {
                    label: "Defined network",
                    value: "DefinedNetwork",
                    description: "Attach to a defined network",
                  },
                  {
                    label: "Host bridge",
                    value: "Bridge",
                    description: "Attach to an host's bridge",
                  },
                ]}
              />
            ) : (
              p.network.type
            )
          }
        />
      </ListItem>
      <div style={{ marginLeft: "70px" }}>
        {/* MAC address input */}
        <MACInput
          editable={p.editable}
          label="MAC Address"
          value={p.network.mac}
          onValueChange={(v) => {
            p.network.mac = v!;
            p.onChange?.();
          }}
        />

        {/* NIC model */}
        <SelectInput
          editable={p.editable}
          label="NIC Model"
          value={p.network.model}
          onValueChange={(v) => {
            p.network.model = v as any;
            p.onChange?.();
          }}
          options={[
            { label: "e1000", value: "E1000" },
            {
              label: "virtio",
              value: "Virtio",
              description:
                "Recommended model, but will require specific drivers on OS that do not support it.",
            },
          ]}
        />

        {/* Defined network selection */}
        {p.network.type === "DefinedNetwork" && (
          <SelectInput
            editable={p.editable}
            label="Defined network"
            options={p.networksList.map((n) => {
              const chars = [n.forward_mode.toString()];
              if (n.ip_v4) chars.push("IPv4");
              if (n.ip_v6) chars.push("IPv6");
              if (n.description) chars.push(n.description);

              return {
                label: n.name,
                value: n.name,
                description: chars.join(" - "),
              };
            })}
            value={p.network.network}
            onValueChange={(v) => {
              if (p.network.type === "DefinedNetwork")
                p.network.network = v as any;
              p.onChange?.();
            }}
          />
        )}

        {/* Bridge selection */}
        {p.network.type === "Bridge" && (
          <SelectInput
            editable={p.editable}
            label="Host bridge"
            options={p.bridgesList.map((n) => {
              return {
                label: n,
                value: n,
              };
            })}
            value={p.network.bridge}
            onValueChange={(v) => {
              if (p.network.type === "Bridge") p.network.bridge = v as any;
              p.onChange?.();
            }}
          />
        )}

        {p.network.type !== "UserspaceSLIRPStack" && (
          <>
            {/* Network Filter */}
            <NWFilterSelectInput
              editable={p.editable}
              label="Network filter"
              value={p.network.nwfilterref?.name}
              onChange={(v) => {
                if (v && !p.network.nwfilterref) {
                  p.network.nwfilterref = { name: v, parameters: [] };
                } else if (v) {
                  p.network.nwfilterref!.name = v;
                } else {
                  p.network.nwfilterref = undefined;
                }
                p.onChange?.();
              }}
              canBeNull={true}
              nwfilters={p.networkFiltersList}
            />

            {p.network.nwfilterref && (
              <div style={{ margin: "10px" }}>
                <VMNetworkFilterParameters
                  filterref={p.network.nwfilterref}
                  {...p}
                />
              </div>
            )}
          </>
        )}
      </div>
    </>
  );
}
