import { Button } from "@mui/material";
import Grid from "@mui/material/Grid";
import React, { ReactElement } from "react";
import { useNavigate } from "react-router-dom";
import {
  NWFilter,
  NWFilterApi,
  NWFilterIsBuiltin,
} from "../../api/NWFilterApi";
import { ServerApi } from "../../api/ServerApi";
import { useAlert } from "../../hooks/providers/AlertDialogProvider";
import { useConfirm } from "../../hooks/providers/ConfirmDialogProvider";
import { useSnackbar } from "../../hooks/providers/SnackbarProvider";
import { AsyncWidget } from "../AsyncWidget";
import { TabsWidget } from "../TabsWidget";
import { XMLAsyncWidget } from "../XMLWidget";
import { EditSection } from "../forms/EditSection";
import { NWFSelectReferencedFilters } from "../forms/NWFSelectReferencedFilters";
import { NWFilterPriorityInput } from "../forms/NWFilterPriorityInput";
import { NWFilterRules } from "../forms/NWFilterRules";
import { SelectInput } from "../forms/SelectInput";
import { TextInput } from "../forms/TextInput";

interface DetailsProps {
  nwfilter: NWFilter;
  editable: boolean;
  onChange?: () => void;
}

export function NWFilterDetails(p: DetailsProps): ReactElement {
  const [nwFiltersList, setNwFiltersList] = React.useState<
    NWFilter[] | undefined
  >();

  const load = async () => {
    setNwFiltersList(await NWFilterApi.GetList());
  };

  return (
    <AsyncWidget
      loadKey={p.nwfilter.uuid}
      load={load}
      errMsg="Failed to load the list of network filters!"
      build={() => (
        <NetworkFilterDetailsInner nwFiltersList={nwFiltersList!} {...p} />
      )}
    />
  );
}

type InnerDetailsProps = DetailsProps & { nwFiltersList: NWFilter[] };

enum NetFilterTab {
  General = 0,
  Rules,
  XML,
  Danger,
}

export function NetworkFilterDetailsInner(
  p: InnerDetailsProps
): React.ReactElement {
  const [currTab, setCurrTab] = React.useState(NetFilterTab.General);

  return (
    <>
      <TabsWidget
        currTab={currTab}
        onTabChange={setCurrTab}
        options={[
          { label: "General", value: NetFilterTab.General, visible: true },
          {
            label: "Rules",
            value: NetFilterTab.Rules,
            visible: p.editable || p.nwfilter.rules.length > 0,
          },

          {
            label: "XML",
            value: NetFilterTab.XML,
            visible: !p.editable,
          },
          {
            label: "Danger zone",
            value: NetFilterTab.Danger,
            color: "red",
            visible: !p.editable && !NWFilterIsBuiltin(p.nwfilter),
          },
        ]}
      />

      {currTab === NetFilterTab.General && (
        <NetworkFilterDetailsTabGeneral {...p} />
      )}
      {currTab === NetFilterTab.Rules && (
        <NetworkFilterDetailsTabRules {...p} />
      )}
      {currTab === NetFilterTab.XML && <NetworkFilterDetailsTabXML {...p} />}
      {currTab === NetFilterTab.Danger && (
        <NetworkFilterDetailsTabDanger {...p} />
      )}
    </>
  );
}

function NetworkFilterDetailsTabGeneral(
  p: InnerDetailsProps
): React.ReactElement {
  return (
    <Grid container spacing={2}>
      {/* Metadata section */}
      <EditSection title="Metadata">
        <TextInput
          label="Name"
          editable={p.editable}
          value={p.nwfilter.name}
          onValueChange={(v) => {
            p.nwfilter.name = v ?? "";
            p.onChange?.();
          }}
          checkValue={(v) => /^[a-zA-Z0-9_-]+$/.test(v)}
          size={ServerApi.Config.constraints.nwfilter_name_size}
        />

        <TextInput label="UUID" editable={false} value={p.nwfilter.uuid} />

        <SelectInput
          label="Chain"
          editable={p.editable}
          value={p.nwfilter.chain?.protocol}
          onValueChange={(v) => {
            p.nwfilter.chain = v ? { protocol: v } : undefined;
            p.onChange?.();
          }}
          options={ServerApi.Config.nwfilter_chains.map((c) => {
            return { label: c, value: c };
          })}
        />

        <NWFilterPriorityInput
          {...p}
          label="Priority"
          value={p.nwfilter.priority}
          onChange={(pri) => {
            p.nwfilter.priority = pri;
            p.onChange?.();
          }}
        />
      </EditSection>

      {/* Referenced filters */}
      {(p.editable || p.nwfilter.join_filters.length > 0) && (
        <EditSection title="Referenced filters">
          <NWFSelectReferencedFilters
            selected={p.nwfilter.join_filters}
            excludedFilters={[p.nwfilter.name]}
            {...p}
          />
        </EditSection>
      )}
    </Grid>
  );
}

function NetworkFilterDetailsTabRules(
  p: InnerDetailsProps
): React.ReactElement {
  return (
    <NWFilterRules
      editable={p.editable}
      rules={p.nwfilter.rules}
      onChange={p.onChange}
    />
  );
}

function NetworkFilterDetailsTabXML(p: InnerDetailsProps): React.ReactElement {
  return (
    <XMLAsyncWidget
      errMsg="Failed to load network filter XML definition!"
      identifier={p.nwfilter.uuid!}
      load={() => NWFilterApi.GetSingleXML(p.nwfilter.uuid!)}
    />
  );
}

function NetworkFilterDetailsTabDanger(
  p: InnerDetailsProps
): React.ReactElement {
  const confirm = useConfirm();
  const snackbar = useSnackbar();
  const alert = useAlert();
  const navigate = useNavigate();

  const requestDelete = async () => {
    try {
      if (
        !(await confirm(
          "Do you really want to delete this network filter?",
          `Delete network filter ${p.nwfilter.name}`,
          "Delete"
        ))
      )
        return;

      await NWFilterApi.Delete(p.nwfilter);

      navigate("/nwfilter");
      snackbar("The network filter was successfully deleted!");
    } catch (e) {
      console.error(e);
      alert(`Failed to delete the network filter!\n${e}`);
    }
  };

  return (
    <Button color="error" onClick={requestDelete}>
      Delete this network filter
    </Button>
  );
}
